#include "Sudoku.h"
#include "SudokuCell.h"
#include "SudokuSound.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <FL/x.H>
#include <FL/Enumerations.H>
#include <FL/Fl_Button.H>
#include <FL/fl_ask.H>
#include <FL/fl_draw.H>
#include <FL/math.h>
#include <FL/Fl_File_Chooser.H>

#include <FL/Fl_Group.H>
#include <FL/Fl_Sys_Menu_Bar.H>
#include <FL/Fl_Help_Dialog.H>

//file IO
#include <iostream>
#include <fstream>
#include <cstring>

#ifdef WIN32
#  include "sudokurc.h"
#elif !defined(__APPLE__)
#  include "sudoku.xbm"
#endif // WIN32

using namespace std;

// ---------------------------------------------------------------------------
// Sudoku class globals...
Fl_Help_Dialog	*Sudoku::help_dialog_ = (Fl_Help_Dialog *)0;
Fl_Preferences	Sudoku::prefs_(Fl_Preferences::USER, "fltk.org", "Sudoku");

// ---------------------------------------------------------------------------
// Create a Sudoku game window...
Sudoku::Sudoku()
    : Fl_Window(GROUP_SIZE * 3, GROUP_SIZE * 3 + MENU_OFFSET, "Sudoku")
{
  solvechecks = 0;  //init check counter
  Fl_Group *g;
  SudokuCell *cell;

  static Fl_Menu_Item	items[] =
    {
      { "&Game", 0, 0, 0, FL_SUBMENU
      },
      { "&New Puzzle", FL_COMMAND | 'n', new_cb, 0, FL_MENU_DIVIDER },

      { "&Check Game", FL_COMMAND | 'c', check_cb, 0, 0 },
      { "&Restart Game", FL_COMMAND | 'r', restart_cb, 0, 0 },
      { "&Show solution", FL_COMMAND | 's', showSolution_cb, 0, FL_MENU_DIVIDER },

      { "&Input Puzzle", FL_COMMAND | 'i', input_cb, 0, 0},
      { "Save Puzzle", FL_COMMAND | 'x', savepuzzle_cb, 0, 0}, //e&Xport
      { "S&ave Game", FL_COMMAND | 'a', savegame_cb, 0, 0},
      { "&Load Game", FL_COMMAND | 'l', loadgame_cb, 0, FL_MENU_DIVIDER },
      
      { "Save &offline Game", FL_COMMAND | 'o' , printgame_cb, 0, FL_MENU_DIVIDER },
      
      { "&Quit", FL_COMMAND | 'q', close_cb, 0, 0 },
      { 0 },
      { "&Difficulty", 0, 0, 0, FL_SUBMENU },
      { "&Easy(50)", 0, diff_cb, (void *)"0", FL_MENU_RADIO },
      { "&Medium(40)", 0, diff_cb, (void *)"1", FL_MENU_RADIO },
      { "&Hard(30)", 0, diff_cb, (void *)"2", FL_MENU_RADIO },
      { "&Impossible(20)", 0, diff_cb, (void *)"3", FL_MENU_RADIO },
      { 0 },
      { "&Functions", 0, 0, 0, FL_SUBMENU },
      { "Has &Sound", FL_ALT | 's' , hassound_cb, 0, FL_MENU_TOGGLE },
      { "&Hints", FL_COMMAND | 'h', hints_cb, 0,0 },
      { "&Hint This cell", FL_ALT | 'h', hintthis_cb, 0,0 },
      { "&Update Hints from this cell", FL_COMMAND | 'u', updatehintsfromthis_cb, 0,0 },
      { "Auto hints update", 0 , autohints_cb, 0, FL_MENU_TOGGLE },
      { "Show unique hints", FL_COMMAND | '1', showuniquehints_cb, 0,0 },
      { "Auto unique hints", 0 , autounique_cb, 0, FL_MENU_TOGGLE },
      { "&Promote unique to values", FL_COMMAND | 'p', promoteunique_cb, 0,0 },
      { "Auto promote unique to values", 0 , autopromoteunique_cb, 0, FL_MENU_TOGGLE },
      { "&Remove Hints", FL_ALT | 'r', removehints_cb, 0,0 },
      { "Show this value", 'v', showthisvalue_cb, 0,0 },
      { "Remove background colors", ' ', removebckgrnds_cb, 0,0 },
      { "Help me: spoil a cell", '?', helpme_cb, 0,0 },
      { "Calculate number of solutions", 0, NumberOfSolutions_cb, 0,0 },
      { 0 },
      { "Help", 0, 0, 0, FL_SUBMENU },  //NO &Help (h shortcut key because h = cell hint)
      { "&About Sudoku", FL_F + 1, help_cb, 0, 0 },
      { 0 },
      { 0 }
    };

  // Setup sound output...
  sound_ = new SudokuSound();

  // Menubar...
  prefs_.get("difficulty", difficulty_, 0);
  prefs_.get("hassound", hassound_, 1);  //1 is default
  sound_->soundon(hassound_);
  prefs_.get("autohints", autohints_, 0);  //0 is default
  prefs_.get("autounique", autounique_, 0);  //0 is default
  prefs_.get("promoteunique", promoteunique_, 0);  //0 is default
  prefs_.get("hassolution", hassolution_, 1);  //1 is default
  prefs_.get("imported", imported_, 0);  //0 is default

  if (difficulty_ < 0 || difficulty_ > 3) difficulty_ = 0;
  //set current difficulty radio "dot" in menu  12
  items[12 +  difficulty_].flags |= FL_MENU_VALUE;

  //set sound toggle  18
  if (hassound_) items[18].flags |= FL_MENU_VALUE;

  //PLJ set autohint toggle 22
  if (autohints_) items[22].flags |= FL_MENU_VALUE;

  //PLJ set autounique toggle 24
  if (autounique_) items[24].flags |= FL_MENU_VALUE;

  //PLJ set autounique toggle 26
  if (promoteunique_) items[26].flags |= FL_MENU_VALUE;

  menubar_ = new Fl_Sys_Menu_Bar(0, 0, 3 * GROUP_SIZE, 25);
  menubar_->menu(items);

  // Create the grids...
  grid_ = new Fl_Group(0, MENU_OFFSET, 3 * GROUP_SIZE, 3 * GROUP_SIZE);

  for (int i = 0; i < 3; i ++)
  {
    for (int j = 0; j < 3; j ++)
    {
      g = new Fl_Group(j * GROUP_SIZE,
                       i * GROUP_SIZE + MENU_OFFSET,
                       GROUP_SIZE,
                       GROUP_SIZE);
      g->box(FL_BORDER_BOX);
      if ((i == 1) ^ (j == 1))
        g->color(FL_DARK3);
      else
        g->color(FL_DARK2);
      g->end();

      grid_groups_[i][j] = g;
    }
  }

  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      cell = new SudokuCell(j * CELL_SIZE + CELL_OFFSET + (j / 3) * (GROUP_SIZE - 3 * CELL_SIZE),
                            i * CELL_SIZE + CELL_OFFSET + MENU_OFFSET + (i / 3) * (GROUP_SIZE - 3 * CELL_SIZE),
                            CELL_SIZE, CELL_SIZE);

      cell->callback(cell_callback);
      cell->init(i,j);  //each cell gets to know its place, ??to constructor??
      grid_cells_[i][j] = cell;
    }
  }
  // Set icon for window (MacOS uses app bundle for icon...)
#ifdef WIN32
  icon((char *)LoadIcon(fl_display, MAKEINTRESOURCE(IDI_ICON)));
#elif !defined(__APPLE__)
  fl_open_display();
  icon((char *)XCreateBitmapFromData(fl_display, DefaultRootWindow(fl_display),
                                     (char *)sudoku_bits, sudoku_width,
                                     sudoku_height)
      );
#endif // WIN32

  // Catch window close events...
  callback(close_cb);

  // Make the window resizable...
  resizable(grid_);
  size_range(3 * GROUP_SIZE, 3 * GROUP_SIZE + MENU_OFFSET, 0, 0, 5, 5, 1);

  // Restore the previous window dimensions...
  int X, Y, W, H;

  if (prefs_.get("x", X, -1))
  {
    prefs_.get("y", Y, -1);
    prefs_.get("width", W, 3 * GROUP_SIZE);
    prefs_.get("height", H, 3 * GROUP_SIZE + MENU_OFFSET);

    resize(X, X, W, H);

    lastFocusI = -1; //init as no focus yet
    lastFocusJ = -1;
  }

  set_title();

  import_stopped = false;
}// constructor Sudoku::Sudoku()

// Destroy the sudoku window...
Sudoku::~Sudoku()
{
  delete sound_;
}// destructor Sudoku::~Sudoku()

// ---_cb---------------------------------------------------------------------
// Check for a solution or wrong values to the game...
void
Sudoku::check_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->cleanBackgrounds(); //remove old backgrounds
  s->check_game(true);
  s->redraw();
}//check_cb

// Close the window, saving the game first...
void
Sudoku::close_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->save_pref_game();
  s->hide();
  if (help_dialog_) help_dialog_->hide();
}// close_cb

// Set the level of difficulty...
void
Sudoku::diff_cb(Fl_Widget *widget, void *d)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int diff = atoi((char *)d);

//prev was imported or a different difficulty: so do it
  if ((s->imported_) ||(diff != s->difficulty_))
  {
    s->difficulty_ = diff;

    //switch back from imported files
    s->imported_ = 0;
    prefs_.set("imported", s->imported_);
    s->hassolution_ = 1;
    prefs_.set("hassolution", s->hassolution_);

    s->new_game(s->seed_);
    s->set_title();

    prefs_.set("difficulty", s->difficulty_);
    s->redraw();
  }
}// diff_cb

// Show the on-line help...
void
Sudoku::help_cb(Fl_Widget *, void *)
{
  if (!help_dialog_)
  {
    help_dialog_ = new Fl_Help_Dialog();

    help_dialog_->value(
      "<HTML>\n"
      "<HEAD>\n"
      "<TITLE>Sudoku Help</TITLE>\n"
      "</HEAD>\n"
      "<BODY BGCOLOR='#ffffff'>\n"
      "Version " VERSION  "-build "   BUILD  "\n"
      "<H2>About the Game</H2>\n"

      "<P>Sudoku (pronounced soo-dough-coo with the emphasis on the\n"
      "first syllable) is a simple number-based puzzle/game played on a\n"
      "9x9 grid that is divided into 3x3 subgrids (boxes). The goal is to enter\n"
      "a number from 1 to 9 in each cell so that each number appears\n"
      "only once in each column and row. In addition, each 3x3 subgrid\n"
      "may also only contain one of each number.</P>\n"

      "<P>This version of the puzzle is Copyright 2006 by Paul Janson &\n"
      "Rob Golsteijn<br>\n"
      "Based on the version from FLTK example by Michael R Sweet Copyright 2005.\n"
      "<P>And modifications by John T. Haller of PortableApps.com to improve\n"
      "portability. (version 1.1.7a)</p>\n"
      
      "<H2>How to Play the Game</H2>\n"

      "<P>At the start of a new game, Sudoku fills in a random selection\n"
      "of cells for you - the number of cells depends on the difficulty\n"
      "level you use. Click in any of the empty cells or use the arrow\n"
      "keys to highlight individual cells and press a number from 1 to 9\n"
      "to fill in the cell. To clear a cell, press 0, Delete, or\n"
      "Backspace. When you have successfully completed all subgrids, the\n"
      "entire puzzle is highlighted in green until you start a new\n"
      "game.</P>\n"

      "<P>As you work to complete the puzzle, you can display hints (possible\n"
      "solutions) inside each cell by holding the Shift key and pressing\n"
      "each number in turn. Repeat the process to remove individual\n"
      "numbers, or press a number without the Shift key to replace them\n"
      "with the actual number to use.	Note that this will only work with\n"
    	"the standard number keys and not the numeric keypad.</P>\n"
      //John's improved help text

      "<P>IN PROGRESS update the manual for the added features!!!!</P>\n"

      "<H2>The Game Window</H2>\n"
      "<P>The window title shows the difficulty level or if the game was\n"
      "imported. The Help menu shows this help file.</P>\n"

      "<H2>Games Menu</H2>\n"
      "<P><b>New Puzzle</b> asks the user if a new puzzle (of the\n"
      "current difficulty level) must be created.</P>\n"
      "<P><b>Check Game</b> checks if the game is solved, and colors every cell\n"
      "GREEN if so. Cells are highlighted YELLOW if conflicting values are\n"
      "found.</P>\n"
      "<P><b>Restart Game</b> removes all user entered values and hints.\n"
      "Hints will be entered immediatly if \"Auto hints update\" is switch\n"
      "on.</P>\n"
      "<P><b>Show Solution</b> will show a solution in DARK YELLOW if a\n"
      "solution is available. Every generated game has (at least) one possible\n"
      "solution. Imported games only have a solution if this is calculated with\n"
      "the \"Calculate number of solutions\" function. A message is shown if no\n"
      "solution is available.</P>\n"
      "<P><b>Input Puzzle</b> lets the user enter a new puzzle in a text input\n"
      "window as a string of numbers and dots or zeros.</P>\n"
      "<P><b>Save Puzzle</b> lets the user save the puzzle as a string in a\n"
      "file (*.sdk). The content of the file can be used as input for the\n"
      "\"Input Puzzle\" function. The file can also be loaded with the \"Load\n"
      "Game\" function.</P>\n"
      "<P><b>Save Game</b> lets the user save the whole game in a file (*.su).\n"
      "A whole game contains the puzzle string and a line for each cell\n"
      "containing the user selected value and hints.</P>\n"
      "<P><b>Load Game</b> lets the user load the whole game from a file (*.su).\n"
      "If only the puzzle (as saved with \"Save Puzzle\" is available the new\n"
      "puzzle will be presented.</P>\n"
      "<P><b>Save offline Game</b> saves the game into a txt file that can be\n"
      "printed for offline playing.</P>\n"
      "<P><b>Quit</b> Quits the Sudoku game which saves the current situation to\n"
      "the preferences. This situation will be loaded when restarted.</P>\n"

      "<H2>Difficulty Menu</H2>\n"
      "<P>The difficulty level of Easy(50), Medium(40), Hard(30) or\n"
      "Impossible(20) can be selected. Which only determines the number of given\n"
      "cells. These generated games have at least one solutions, but more might\n"
      "be possible. The \"Calculate number of solutions\" function can be used\n"
      "to determine how many solutions are there.</P>\n"

      "<H2>Functions Menu</H2>\n"
      "<P><b>Has Sound</b> Toggles the sound on or off.\n"
      "Sound off speeds up the animations.</P>\n"
      "<P><b>Hints</b> calculates the possible solutions (hints) for each empty\n"
      "cell. Hints are calculated based on the row, column and box. This removes\n"
      "the initial work of filling in all 9 hints and then removing row, column\n"
      "and box values. Hints in a RED font and on a fixed position for each\n"
      "value to create an easy overview.</P>\n"
      "<P><b>Hint This cell</b> calculates the hints only for this cell.\n"
      "Hints are calculated based on the row, column and box.</P>\n"
      "<P><b>Update Hints from this cell</b> updates hints of other cells\n"
      "based on the value of this cell. It removes this value from the hints in"
      "cells in it's row, column and box.</P>\n"
      "<P><b>Auto hints update</b> Toggles the auto hints update on or off.\n"
      "This allows the hints in the game to be automaticly updated if a value\n"
      "is entered in a cell.</P>\n"
      "<P><b>Show unique hints</b> highlights all cells CYAN with unique hints.\n"
      "Hints are unique within a row, column, box or the only hint left for this\n"
      "cell. The unique hint value is in a BLACK font. This is only usefull if\n"
      "all hints are kept from the beginning otherwise hints might have become\n"
      "unique by lack of other hints, which results in a wrong conclusion of\n"
      "being unique.</P>\n"
      "<P><b>Auto unique hints</b> Toggles the auto unique hints on or off.\n"
      "This autmaticly shows the unique hints when a value is entered.\n"
      "This is only useful if auto hints update is also switch on.</P>\n"
      "<P><b>Promote unique to values</b> promotes unique hints to values.\n"
      "This is only usefull if all hints are kept from the beginning\n"
      "otherwise hints might have become unique by lack of other hints,\n"
      "which renders the promoted value untrusted.</P>\n"
      "<P><b>Auto promote unique to value</b> Toggles the auto promote unique\n"
      "to value on or off. This function will automaticly promote unique hints.\n"
      "Uniqueness will not be visable anymore, because the cell gets its value\n"
      "imediately.</P>\n"
      "<P><b>Remove Hints</b> removes the hints from all cell showing them.</P>\n"
      "<P><b>Show this value</b> highlights all cells with the same value as\n"
      "the focus cell MAGENTA. If selected for another value the previous\n"
      "magenta cell backgrounds will be reset.</P>\n"
      "<P><b>Remove background colors</b> removes all colors given by the\n"
      "support functions (YELLOW check warning, CYAN unique hint, MAGENTA\n"
      " this value, GREEN correct/spoiled cell, DARK YELLOW solution).</P>\n"
      "<P><b>Help me: spoil a cell</b> will randomly pick an empty cell and fill\n"
      "it with the solution value and highlight it green. If there are more\n"
      "possible solutions for this puzzle this value might be a conflicting\n"
      "with already available user enter values, working towards a different\n"
      "solution.</P>\n"
      "<P><b>Calculate number of solutions</b> calculates the number of\n"
      "solutions using backtracking. This might take a long time for an empty\n"
      "game. The first solution found will be save into the solution of the game\n"
      "if the game didn't have a solution yet (as is the case for imported\n"
      "games). A dialog will tell if there was one or more solutions (it stops\n"
      "at second found) and the time it took to calculate until it stops.</P>\n"
      "</BODY>\n"
    );
  }
  help_dialog_->show();
}//help_cb

// Create a new game...
void
Sudoku::new_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);

  if (s->grid_cells_[0][0]->color() != FL_GREEN)
  {
    if (!fl_choice("Are you sure you want to change the difficulty level and "
                   "discard the current game?", "Keep Current Game", "Start New Game",
                   NULL)) return;
  }
  //reset the title after an imported game
  s->imported_ = 0;
  s->set_title();

  s->new_game(time(NULL));
  s->redraw();
}//new_cb

// Restart game from beginning... only show readonly() cells
void
Sudoku::restart_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window());
  s->cleanBackgrounds();

  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      SudokuCell *cell = s->grid_cells_[i][j];

      if (!cell->readonlyGet())
      {
        int v = cell->valueGet();
        cell->valueSet(0);
        cell->color(FL_LIGHT3);

        if (v)
        {
          cell->redraw();
          s->sound_->play('A' + v - 1);
        }
      }//use cell
    }//j
  }//i

  //Handle options aspect
  if ( s->getautohints() ) s->hints();
  if ( s->getautounique() ) s->showuniquehints();
  if ( s->getpromoteunique() ) s->promoteuniquehints();
  s->check_game(false); //check if game is OK->green
  s->redraw();
}//restart_cb

// show Solution of the puzzle...
void
Sudoku::showSolution_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);

  s->cleanBackgrounds();
  s->showSolution_game();

  s->redraw();
}//showSolution_cb

// Input a puzzle as string...
void
Sudoku::input_cb(Fl_Widget *widget, void *)
{

  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  const char * import = fl_input("new sudoku:", DEFAULTPUZZLE);
  if (import != NULL)
  {
    printf("import=%s\n", import);
    s->input_game((char *)import);

    s->imported_ = 1; //imported
    prefs_.set("imported", s->imported_);
    s->hassolution_ = 0; //imported
    prefs_.set("hassolution", s->hassolution_);
    s->set_title();
    
    //Handle options aspect
    if ( s->getautohints() ) s->hints();
    if ( s->getautounique() ) s->showuniquehints();
    if ( s->getpromoteunique() ) s->promoteuniquehints();
    s->check_game(false); //check if game is OK->green
    s->redraw();
  }
}//input_cb

// Save the puzzle (readonly cells) to a file .sdk
void
Sudoku::savepuzzle_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);

  char* filename = fl_file_chooser("save puzzle", "*.sdk", ".", 0);
  //TODO check for extension sdk and add if no extension
  ofstream myfile;
  myfile.open (filename);

  if (myfile.is_open())
  {
    for (int i=0; i<9;i++)
      for (int j=0;j<9;j++)
      {
        SudokuCell *tempcell;
        tempcell = s->grid_cells_[i][j];
        if (tempcell->readonlyGet())
        {
          myfile << tempcell->valueGet();
        }
        else
        {
          myfile << ".";
        }
      }
    myfile.close();
  }//if open
}// savepuzzle_cb

// Save a game (puzzle including the usert values and hints) to a file .su
void
Sudoku::savegame_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);

  char* filename = fl_file_chooser("save game", "*.su", ".", 0);
  //TODO check for extension su and add if no extension
  ofstream myfile;
  myfile.open (filename);
  if (myfile.is_open())
  {
    for (int i=0; i<9;i++)
      for (int j=0;j<9;j++)
      {
        SudokuCell *tempcell;
        tempcell = s->grid_cells_[i][j];
        if (tempcell->readonlyGet())
        {
          myfile << tempcell->valueGet();
        }
        else
        {
          myfile << ".";
        }
      }

    myfile << "\n";

    //cell per line value, hints list; value = 0 if no value
    for (int i=0; i<9;i++)
    {
      for (int j=0;j<9;j++)
      {
        SudokuCell *tempcell;
        tempcell = s->grid_cells_[i][j];
        myfile << i << "," << j << ":";

        myfile << tempcell->valueGet();
        for (int h = 1; h <=9 ; h++)
        {
          int hint =  tempcell->hintsGet(h);
          if (hint) myfile << hint;
        }
        myfile << "\n";
      }
    }
    myfile.close();
  }// if open
}// savegame_cb

// Load a saved game (puzzle including the usert values and hints) to a file .su
// If only the first line is used it loads a puzzle (as saved in a *.sdk)
void
Sudoku::loadgame_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  SudokuCell *tempcell;
  const int MAXLINESIZE = 85;
  char line[MAXLINESIZE];
  int index = 0;
  int i,j;

  char* filename = fl_file_chooser("load game", "*.su", ".", 0); //relative = 0

  //if file exist start reading
  //FILE IO based on http://www.cplusplus.com/doc/language/tutorial/files.html ,Thanx guys
  ifstream myfile (filename);
  if (myfile.is_open())
  {
    //clean up all cell to make a new puzzle:
    for (int ii = 0; ii < 9; ii ++)
    {
      for (int jj = 0; jj < 9; jj ++)
      {
        SudokuCell * cell = s->grid_cells_[ii][jj];

        cell->valueSet(0);
        cell->readonlySet(0);
        cell->color(FL_LIGHT3);
        cell->resetHints();
      }
    }//clean up existing cells

    //parse the first line of the file
    //This means : Get the Puzzle
    if (! myfile.eof() )
    {
      myfile.getline (line,MAXLINESIZE);

      index = 0;
      for (int ii=0; ii<9;ii++)
      {
        for (int jj=0;jj<9;jj++)
        {
          int value =  line[index] - '0';

          tempcell = s->grid_cells_[ii][jj];

          if ( value > 0 && value <= 9 )
          {
            tempcell->readonlySet(1);
            tempcell->valueSet(value);
            tempcell->color(FL_GRAY);
          }
          index++;
        }// for jj
      }//for ii
    }//not eof

    // get user values and hints if they are available
    index = 0;
    bool hasHints = false;

    while (! myfile.eof() && index < 81)
    {
      myfile.getline (line,MAXLINESIZE);
      //typical line format with value: 0,5:9      //i,j:value
      //typical line format with hints: 5,2:02458  //i,j:0hints
      i = line[0] - '0';
      j = line[2] - '0';
      tempcell = s->grid_cells_[i][j];

      int value = line[4] - '0';
      if (value != 0)
      {
        //if not a Puzzle value set the user value
        if (!tempcell->readonlyGet())
        {
          tempcell->valueSet(value);
          //printf("USER value = %d\n",value);
        }
        //else{printf("PUZZLE value = %d\n", value); }
      }
      else
      {
        int linesize = strlen ( line );
        if (linesize >= 5)
        {
          for (int teller = 5; teller < linesize; teller++)
          {
            if (!hasHints) hasHints=true;
            value = line[teller]-'0';
            tempcell->hintsSet(value);
          }
        }
      }
      index++;
    }

    myfile.close();

    //set prefs_ to imported
    s->imported_ = 1;
    prefs_.set("imported", s->imported_);
    s->hassolution_ = 0;
    prefs_.set("hassolution", s->hassolution_);

    //NO user imports a specific game so lets continue as saved !!
    s->redraw();
  }
  else
  {
    //cout << "Unable to open file (" << filename << ")\n";
    char msgbuf [255];
    sprintf(msgbuf,"Unable to open file (%s)\n", filename);
    fl_message(msgbuf);
  }
}// loadgame_cb

// Toggle the sound setting...
void
Sudoku::hassound_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int hassound = ((Fl_Menu_ *)widget)->mvalue()->value() ==  FL_MENU_VALUE;

  if (hassound != s->hassound_)
  {
    s->hassound_ = hassound;
    prefs_.set("hassound", s->hassound_);
    s->sound_->soundon(hassound);
  }
}// hassound_cb

// Toggle the autohints setting...
void
Sudoku::autohints_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int autohints = ((Fl_Menu_ *)widget)->mvalue()->value() ==  FL_MENU_VALUE;

  if (autohints != s->autohints_)
  {
    s->autohints_ = autohints;
    prefs_.set("autohints", s->autohints_);
  }
}// autohints_cb

// Toggle the auto unique hints setting...
void
Sudoku::autounique_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int autounique = ((Fl_Menu_ *)widget)->mvalue()->value() ==  FL_MENU_VALUE;

  if (autounique != s->autounique_)
  {
    s->autounique_ = autounique;
    prefs_.set("autounique", s->autounique_);
  }
}// autounique_cb

// Toggle the auto promot unique hints to values setting...
void
Sudoku::autopromoteunique_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int promoteunique = ((Fl_Menu_ *)widget)->mvalue()->value() ==  FL_MENU_VALUE;

  if (promoteunique != s->promoteunique_)
  {
    s->promoteunique_ = promoteunique;
    prefs_.set("promoteunique", s->promoteunique_);
  }
}//autopromoteunique_cb

// Fill all cells with hints, calculate for being unique in roq/column/box ...
void
Sudoku::hints_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->hints();
  s->redraw();
}// hints_cb

// Remove all hints from the game ...
void
Sudoku::removehints_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->removehints();
  s->redraw();
}// removehints_cb

// Remove the background colors ...
void
Sudoku::removebckgrnds_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->cleanBackgrounds();
  s->redraw();
}// removebckgrnds_cb

// Calculate the hints for this cell only ...
// as a backup when a user made a type and filed in a value instead of a hint
void
Sudoku::hintthis_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int i = s->lastFocusI;
  int j = s->lastFocusJ;
  s->hintthiscell(i,j);
  widget->redraw(); //redraw cell only!
}//hintthis_cb

// Update hints in other cells based on the value in this cell
void
Sudoku::updatehintsfromthis_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int i = s->lastFocusI;
  int j = s->lastFocusJ;
  s-> updatehintsfromthis(i,j);
  s->redraw();
}//updatehintsfromthis_cb

// Show other cells with the same value as this one
void
Sudoku::showthisvalue_cb(Fl_Widget *widget, void *)
{
  //showsall cells with this value on the grid
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  int i = s->lastFocusI;
  int j = s->lastFocusJ;
  s-> showthisvalue(i,j);
  s->redraw();
}//showthisvalue_cb

// Show all unique hints ...
// draw the unique hint black (in stead of red) and highlight all cells
void
Sudoku::showuniquehints_cb(Fl_Widget *widget, void *)
{
  // loop rows and column and box to remove this value
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->showuniquehints();
  if (s->getpromoteunique()) s->promoteuniquehints();
  s->check_game(false); //check if game is OK->green
  s->redraw();
}//showuniquehints_cb

// Promote all unique hints into values ...
void
Sudoku::promoteunique_cb(Fl_Widget *widget, void *)
{
  // loop rows and column and box to remove this value
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->promoteuniquehints();
  s->check_game(false); //check if game is OK->green
  s->redraw();

}//promoteunique_cb

// Calculate the number of solutions give the current game
// (puzzle as well as user values) use backtracking
void
Sudoku::NumberOfSolutions_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->numberOfSolutions();
  s->redraw();
}//NumberOfSolutions_cb

// Gives a real value, if a solutions is available
// TODO a problem exists for sudoku puzzles with multiple solutions,
// the given value could be offending to user filled in values!!
void
Sudoku::helpme_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->helpme();
  s->redraw();
}//helpme_cb

// save to file a printable the game in different ways
void
Sudoku::printgame_cb(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  char* filename = fl_file_chooser("save printable game", "*.txt", ".", 0);
  s->printGame(filename);
}//printgame_cb
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//A callback routine, as return from a cell to the window (only after a user action)
void
Sudoku::cell_callback(Fl_Widget *widget, void *)
{
  Sudoku *s = (Sudoku *)(widget->window() ? widget->window() : widget);
  s->check_game(false); //check if this game is already solved -> makes it green:)
  s->redraw();
}//cell_callback

// ---Game methods------------------------------------------------------------
// Check if the user has correctly solved the game...
// Highlight YELLOW offending cells (user cells only)
// Or for a correct  sudoku all cells GREEN
void
Sudoku::check_game(bool highlight)
{
  bool empty = false;
  bool correct = true;
  bool completed = true;
  int i, j, k;

  //are we completely filled in
  for (i = 0; i < 9 && completed; i ++)
  {
    for (j = 0; j < 9 && completed; j ++)
    {
      //for each cell do:
      SudokuCell *cell = grid_cells_[i][j];
      int val = cell->valueGet();
      if (!val)
      {
        completed = false;
      }
    }
  }

  // Check the game for right/wrong answers...
  for (i = 0; i < 9; i ++)
  {
    for (j = 0; j < 9; j ++)
    {
      //for each cell do:

      SudokuCell *cell = grid_cells_[i][j];
      int val = cell->valueGet();

      //skip puzzle cells
      if (cell->readonlyGet()) continue;

      //all cells left are user or empty cells
      if (!val)
        empty = true;
      else
      {
        //loop row/col/box
        for (k = 0; k < 9; k ++) //UPDATED added box checks here
        {
          int boxi = (i/3)*3 + (k%3);
          int boxj = (j/3)*3 + k/3;

          if ((i != k && grid_cells_[k][j]->valueGet() == val) ||
              (j != k && grid_cells_[i][k]->valueGet() == val) ||
              (boxi != i && boxj != j && grid_cells_[boxi][boxj]->valueGet() == val)
             ) break;

        }
        if (k < 9)
        {
          //break early so duplicate found
          if (highlight || completed)
            //TODO opgeloste puzzle groen worden of check laten zien
          {
            cell->color(FL_YELLOW);
            cell->redraw();
          }
          correct = false;
        }
        else if (highlight)
        {
          //no duplicate found
          cell->color(FL_LIGHT3);
          cell->redraw();
        }
      }//val
    }//j
  }//i

  if (!empty && correct)
  {
    // Success! make everything GREEN
    for (i = 0; i < 9; i ++)
    {
      for (j = 0; j < 9; j ++)
      {
        SudokuCell *cell = grid_cells_[i][j];
        cell->color(FL_GREEN);
        //cell->readonlySet(1); PLJ delibarately removed
      }
      sound_->play('A' + grid_cells_[i][8]->valueGet() - 1);
    }
  }
}//check_game

// Create a new game...
//  no garantee about unique solutions not even in medium (see or counter)
void
Sudoku::new_game(time_t seed)
{
  int i, j, k, m, t, count;

  // Generate a new (valid) Sudoku grid...
  seed_ = seed;
  srand(seed);

  memset(grid_values_, 0, sizeof(grid_values_));
  //PLJ reset the hints

  for (i = 0; i < 9; i += 3)
  {
    for (j = 0; j < 9; j += 3)
    {
      for (t = 1; t <= 9; t ++)
      {
        for (count = 0; count < 20; count ++)
        {
          k = i + (rand() % 3);
          m = j + (rand() % 3);
          if (!grid_values_[k][m])
          {
            int kk;

            for (kk = 0; kk < k; kk ++)
              if (grid_values_[kk][m] == t) break;

            if (kk < k) continue;

            int mm;

            for (mm = 0; mm < m; mm ++)
              if (grid_values_[k][mm] == t) break;

            if (mm < m) continue;

            grid_values_[k][m] = t;
            break;
          }
        }

        if (count == 20)
        {
          // Unable to find a valid puzzle so far, so start over...
          j = 9;
          i = -3;
          memset(grid_values_, 0, sizeof(grid_values_));
        }
      }
    }
  }

  // Start by making all cells editable & init them
  SudokuCell *cell;

  for (i = 0; i < 9; i ++)
  {
    for (j = 0; j < 9; j ++)
    {
      cell = grid_cells_[i][j];

      cell->valueSet(0);
      cell->readonlySet(0);
      cell->color(FL_LIGHT3);
      //hints etc. what about hintvalues??
      cell->resetHints();
    }//j
  }//i

  //if sound, we show the animation, draw the cleared cells to start with
  if (hassound_)
  {
    redraw();
  }

  // Show N cells...
  count = 10 * (5 - difficulty_);

  int numbers[9];

  for (i = 0; i < 9; i ++) numbers[i] = i + 1;

  while (count > 0)
  {
    for (i = 0; i < 20; i ++)
    {
      k          = rand() % 9;
      m          = rand() % 9;
      t          = numbers[k];
      numbers[k] = numbers[m];
      numbers[m] = t;
    }

    for (i = 0; count > 0 && i < 9; i ++)
    {
      t = numbers[i];

      for (j = 0; count > 0 && j < 9; j ++)
      {
        cell = grid_cells_[i][j];

        if (grid_values_[i][j] == t && !cell->readonlyGet())
        {
          cell->valueSet(grid_values_[i][j]);
          cell->readonlySet(1);
          cell->color(FL_GRAY);
          //always redraw the cell if sound, to show the animation:
          if (hassound_)
          {
            cell->redraw();
          }
          count --;
          break;
        }//readonly
      }//j
      sound_->play('A' + t - 1);
    }//i
  }//while count

  //if auto update hints run hints
  if (autohints_)
  {
    hints();
  }
}// new_game

// showSolution for this puzzle...
void
Sudoku::showSolution_game()
{
  //check if not in imported mode
  if ( !hassolution_) // if from an imported game do nothing
  {
    char msgbuf [255];
    sprintf(msgbuf,"NO Solution Exists!\nDue to imported game.\nFind a solution using \"Calculate number of solutions\"\n");
    fl_message(msgbuf);
    return;
  }

  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      SudokuCell *cell = grid_cells_[i][j];

      cell->valueSet(grid_values_[i][j]);
      //cell->readonlySet(1);  //PLJ removed deliberately
      if (!cell->readonlyGet())
        cell->color(FL_DARK_YELLOW);
    }//j
    sound_->play('A' + grid_cells_[i][8]->valueGet() - 1);
  }//i
}//showSolution_game

// Input a game from a String (as give by the user in a UI fl_input dialog) in input_cb...
void
Sudoku::input_game(char * import)
{
  memset(grid_values_, 0, sizeof(grid_values_));

  //printf("import=%s\n", import);
  int teller = 0;  //0 tot 80 for grid
  import_stopped = false;

  int value;
  //start looping the grid
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      //read next value for this cell
      value = getNextValue(import, teller);
      SudokuCell *cell = grid_cells_[i][j];


      if ( (value >= 1) && (value <= 9) )
      {
        //store valid value
        grid_values_[i][j]= value;
        cell->valueSet(value);
        cell->readonlySet(1);
        cell->color(FL_GRAY);
      }
      else
      {
        //if invalid create an empty cell 0 / . / EOF
        grid_values_[i][j]= 0;
        cell->valueSet(0);
        cell->readonlySet(0);
        cell->color(FL_LIGHT3);
      }
      teller++;
    }
  }//end loop

  int length = strlen(import);
  if (length < 81)
  {
    //check for TOO SHORT import string
    fl_message("import string is TOO SHORT, padding empty cells!\n");
  }
  else   if (length > 81)
  {
    //check for TOO LONG import string
    fl_message("import string is TOO LONG, ignoring additional characters!!!\n");
  }
}//input_game the puzzle

// Calulate all the hints based on row/column/box for empty cells
void
Sudoku::hints()
{
//start looping the grid
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      //get the cell
      SudokuCell *cell = grid_cells_[i][j];
      //cell->printCell();

      if ( !cell->valueGet() )
      {
        //init assume all hints are possible
        for (int hints=1; hints<=9; hints++)
          cell->hintsSet(hints);

        //strike out all found values from possible hints
        //look in row/col/box
        for (int k=0; k<9; k++) /* loop over 9 values in row, column and box */
        {
          for ( int v=1; v<=9; v++)
          {
            if ( grid_cells_[k][j]->valueGet()                           == v ||
                 grid_cells_[i][k]->valueGet()                           == v ||
                 grid_cells_[(i/3)*3 + (k%3)][(j/3)*3 + k/3]->valueGet() == v
               )
            {
              cell->hintsRemove(v);
            }
          }//v
        }//k
      }//cell with hints
    }//j
  }//i
}//hints()

// Reset the hints collected for empty cells,
// and reset the background colors which also give hints ;^)
void
Sudoku::removehints()
{
  SudokuCell *cell;

//start looping the grid
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      //for each cell

      //get the cell
      cell = grid_cells_[i][j];
      //cleanup the hints
      cell->resetHints();
      //reset the background color
      if (!cell->readonlyGet())
        cell->color(FL_LIGHT3);
      else
        cell->color(FL_GRAY);
    }
  }//end loop
}//removehints

// Calculate the hints for this specific cell
void
Sudoku::hintthiscell(int i, int j)
{
  //get the cell
  SudokuCell *cell = grid_cells_[i][j];

  if ( !cell->valueGet() )//only empty cells
  {
    //initally everything is possible
    for (int hints=1; hints<=9; hints++)
      cell->hintsSet(hints);

    for (int k=0; k<9; k++) /* 9 values in row, column and block */
    {
      for ( int v=1; v<=9; v++)
      {
        if ( grid_cells_[k][j]->valueGet()                           == v || //row
             grid_cells_[i][k]->valueGet()                           == v || //column
             grid_cells_[(i/3)*3 + (k%3)][(j/3)*3 + k/3]->valueGet() == v    //box
           )
        {
          cell->hintsRemove(v);
        }//remove initial hint if found in row colmn or box
      }//v loop for all values
    }//k loop iterator within row/col/box
  }//do open cells only
}//hintthiscell

// Update rows and column and box to remove this value from the hints
void
Sudoku::updatehintsfromthis(int i, int j)
{
  //get the cell
  SudokuCell *cell = grid_cells_[i][j];
  SudokuCell *tempcell;

  int value = cell->valueGet();
  if ( value )//only if this cell has a value
  {
    for (int k=0; k<9; k++) // walk values in row, column and box
    {
      //row
      tempcell = grid_cells_[k][j];
      tempcell->hintsRemove(value);
      //column
      tempcell = grid_cells_[i][k];
      tempcell->hintsRemove(value);
      //box
      tempcell = grid_cells_[(i/3)*3 + (k%3)][(j/3)*3 + k/3];
      tempcell->hintsRemove(value);

    }//k loop iterator within row/col/box
  }//only if the cell has a value
}//updatehintsfromthis

// Highlight (MAGENTA) all cells with this value
void
Sudoku::showthisvalue(int i, int j)
{
  //get the cell
  SudokuCell *cell = grid_cells_[i][j];
  SudokuCell *tempcell;

  int value = cell->valueGet();

  if ( value )//only filled in cells (puzzle aswell as user)
  {
    //loop grid
    for (int i=0; i<9; i++)
    {
      for (int j=0; j<9; j++)
      {
        tempcell = grid_cells_[i][j];
        if ( tempcell->valueGet() == value)
        {
          //found a match
          tempcell->color(FL_MAGENTA);
        }
        else
        {
          //remove highlight of previous value
          if (tempcell->color() == FL_MAGENTA)
          {
            //found old magenta cell
            if ( tempcell->readonlyGet())
            {
              //restore Puzzle cell color
              tempcell->color(FL_GRAY);
            }
            else
            {
              //restore user cell color
              tempcell->color(FL_LIGHT3);
            }
          }
        }
      }//j
    }//i

  }//do open cells only
}//showthisvalue

//remove all background colors user for different hints:
// yellow : check to be offending
// magenta: cells with this value
// cyan   : unique hint
// green  : solution / correct if all cells are green
void
Sudoku::cleanBackgrounds()
{
  SudokuCell *tempcell;
  //loop grid
  for (int i=0; i<9; i++)
  {
    for (int j=0; j<9; j++)
    {
      tempcell = grid_cells_[i][j];
      if ( tempcell->readonlyGet())
      {
        tempcell->color(FL_GRAY);
      }//found a match
      else
      {
        tempcell->color(FL_LIGHT3);
      }
      if (tempcell->uniquehintGet())
      {
        tempcell->uniquehintSet(0);
      }
    }//j
  }//i
}//cleanBackgrounds

// Highlight unique hints
// Highlight background color is CYAN and the unique hint number is black
// unique in a row / column / box or cell (only hint left)
// in a cell only works if we start with 9 hints and remove from there
void
Sudoku::showuniquehints()
{
  //search all rows, columns and boxs for unique hints-> highlight box and hint nr
  //originally only look within the hints now also include the values given
  SudokuCell *tempcell;

  //cleanup the grid first
  // - remove unique hint values
  // - remove tempcell->color(FL_MAGENTA);

  for (int i=0; i<9;i++)
  {
    for (int j=0;j<9;j++)
    {
      tempcell = grid_cells_[i][j];
      if (tempcell->uniquehintGet())
      {
        tempcell->uniquehintSet(0);
        tempcell->color(FL_LIGHT3);
      }
    }//j
  }//i

  //[which row][hint], value = -1 not found yet, value [1,9] first cell with this hint,
  //-2: more than one hint
  int row[9][10];
  int col[9][10];
  int box[9][10];
  //init :)
  memset(row, -1, sizeof(row));
  memset(col, -1, sizeof(col));
  memset(box, -1, sizeof(box));
  memset(cellHint, -1, sizeof(cellHint));

  //loop trough col/row/boxes => collect data
  for (int  z=0; z<9; z++) // loop over all row, column and boxes
  {
    for (int  k=0; k<9; k++) // walk values in row, column and box
    {
      //printf("k(%d)\n", k);
      //row
      tempcell = grid_cells_[z][k];
      testcell(tempcell, row[z], k);

      //column
      tempcell = grid_cells_[k][z];
      testcell(tempcell, col[z], k);

      //box
      tempcell = grid_cells_[(z/3)*3 + (k%3)][(z%3)*3 + k/3];
      testcell(tempcell, box[z], k);

      //cellH already undated 3 times by testcell in the row/col/box loops
    }//k
  }//z

  //loop through collected data and process the data
  for (int z=0; z<9; z++) // loop over all row, column and boxes
  {
    for (int k=1; k<=9; k++) // /loop trough col/row/boxes : inspect uniques
    {
      //if value then found unique in this row/col/box print it :)
      //ROWS
      int rowindex = row[z][k];
      if (rowindex>= 0)  //!-1 & !-2
      {
        //found a unique in this row
        tempcell = grid_cells_[z][rowindex];
        tempcell->uniquehintSet(k);
        tempcell->color(FL_CYAN);
      }
      //COLUMNS
      int colindex = col[z][k];
      if (colindex>= 0)  //!-1 & !-2
      {
        //found a unique in this row
        //printf("col[%d] = cell(%d,%d) hint value %d\n",z, colindex, z, k);
        tempcell = grid_cells_[colindex][z];
        tempcell->uniquehintSet(k);
        tempcell->color(FL_CYAN);
      }
      //BOXES
      int boxindex = box[z][k];
      if (boxindex>= 0)  //!-1 & !-2
      {
        //found a unique in this row
        //printf("box[%d] = cell(%d,%d) hint value %d\n",z, (z/3)*3 + (boxindex%3),(z%3)*3 + boxindex/3, k);
        tempcell = grid_cells_[(z/3)*3 + (boxindex%3)][(z%3)*3 + boxindex/3];
        tempcell->uniquehintSet(k);
        tempcell->color(FL_CYAN);
      }
      //CELL HINTS
      int singlehint = cellHint[z][k-1];
      if (singlehint > 0)  //!-1 & !-2
      {
        //found a unique in this row
        //printf("col[%d] = cell(%d,%d) hint value %d\n",z, colindex, z, k);
        tempcell = grid_cells_[z][k-1];
        tempcell->uniquehintSet(singlehint);
        tempcell->color(FL_CYAN);
      }
    }//k
  }//z
}//showuniquehints

//search all rows, columns and boxs for unique hints-> highlight box and hint nr
void
Sudoku::promoteuniquehints()
{
  SudokuCell *tempcell;

  //find all uniqueHints and promote this hint to value,
  //and cleanup the unique admin
  // - remove uniquehint value
  // - remove tempcell->color(FL_MAGENTA);

  for (int i=0; i<9;i++)
  {
    for (int j=0;j<9;j++)
    {
      tempcell = grid_cells_[i][j];

      if ((tempcell->uniquehintGet()) )
      {
        tempcell->valueSet(tempcell->uniquehintGet());
        tempcell->uniquehintSet(0);
        tempcell->color(FL_LIGHT3);

        if (autohints_) updatehintsfromthis(i, j);
      }//found uniqueHint
    }//j
  }//i
}//promoteuniquehints

// Show random the solution of one of the empty cells
// and make the background color green
void
Sudoku::helpme()
{
  //if solution available only else mesg
  if (!hassolution_)
  {
    char msgbuf [255];
    sprintf(msgbuf,"Help not possible because no solution is available.\nUse \"Calculate number of solutions\" for imported puzzles to create a solution.\n");
    fl_message(msgbuf);
  }
  else
  {
    //Random find a I J that empty
    SudokuCell *tempcell;

    //temporarily store all empty cells with there solution value
    char emptycells[81][3]; //0:i, 1:j, 2: value  //81 at most are not filled in:)
    char emptyIndex = 0;

    //fill array with empty cells and their solution value
    for (int i=0; i<9;i++)
    {
      for (int j=0;j<9;j++)
      {
        tempcell = grid_cells_[i][j];
        if (tempcell->valueGet() ==0)
        {
          //nothing filled in yet
          emptycells[emptyIndex][0]=i;
          emptycells[emptyIndex][1]=j;
          emptycells[emptyIndex][2]=grid_values_[i][j]; //the solution value
          emptyIndex++;
        }
      }// j
    }// i

    if (emptyIndex ==0)
    {
      //no empty cell :(
      char msgbuf [255];
      sprintf(msgbuf,"No cell to help with found, all cells filled in.\nUse \"Check Game\" to find out the result of your work.\n");
      fl_message(msgbuf);
    }
    else
    {
      //Random find number with indexsize (check for no empty cells)
      //time_t tijd = time(NULL); srand(tijd);
      int rndmIdx = rand()%emptyIndex;
      //printf("unknownsIndex= %d & idx=%d\n", emptyIndex, rndmIdx);
      tempcell = grid_cells_[ emptycells[rndmIdx][0] ][ emptycells[rndmIdx][1] ];

      //Show solution as value and change cell background to green
      tempcell->valueSet( emptycells[rndmIdx][2]);
      tempcell->color(FL_GREEN);
    }
  }
}//helpme

//Find how many solutions there are,
// and write the first one to the solution
// if this puzzle didn't have a solution yet, due to imports
void
Sudoku::numberOfSolutions()
{
  SudokuCell *tempcell;
  int solutions = 0; //resulted number of solutions
  int hint;  //temp value for hints
  time_t starttime;
  time_t endtime;
  solvechecks = 0;

  starttime = time (NULL);

  //first Clone the game (puzzle values & user values) into a backtracking testGrid.
  //char testGrid[9][9][10];
  // i  j       values;
  //       [0] the actual value,
  //       [1.9] hints,
  //       [10] bool 1: value / 0: hints
  memset(testGrid, 0, sizeof(testGrid));
  // fill with values
  for (int i = 0; i<9; i++)
  {
    for (int j = 0; j<9; j++)
    {
      tempcell = grid_cells_[i][j];
      testGrid[i][j][0] = tempcell->valueGet();
      if (testGrid[i][j][0]) testGrid[i][j][10] = 1; //1: value 0: hints
    }//j
  }//i

  //And find all possible hints ( copied from hints() )
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      if ( !testGrid[i][j][10] ) //not a value
      {
        // init all hint values as beeing available
        for (hint=1; hint<=9; hint++)
          testGrid[i][j][hint]= hint;

        //remove hints that are in row/col/box
        for (int k=0; k<9; k++) /* 9 values in row, column and block */
        {
          for ( int h=1; h<=9; h++)
          {
            if ( testGrid[k][j][0]                           == h ||
                 testGrid[i][k][0]                           == h ||
                 testGrid[(i/3)*3 + (k%3)][(j/3)*3 + k/3][0] == h
               )
            {
              testGrid[i][j][h]= 0; //remove the hint
            }
          }//h
        }//k
      }//if no value do the hints :)
    }//j
  }//i: end loop
  //TODO possible improvements: by running unique hints & promote,
  //     until there are no unque hints left.

  //start Backtracking (over all hints) until all hints covered
  solutionsStopped = false;  //break off at 2nd solution
  //The recursive Backtracking routine:
  solveGrid( 0, &solutions);

  //do the aftermath: time taken, messages
  endtime = time (NULL);
  long deltatime = endtime - starttime;
  int hours = deltatime / 3600;
  int min = deltatime / 60;
  int sec = deltatime % 60;

  //if solutions = 1 OK, >1 multiple solutions, 0 user made a mistake
  char msgbuf [255];
  if (solutions ==1)
  {
    sprintf(msgbuf,"This is a REAL Sudoku, there is a SINGLE solution!\n Time spend solving= %d:%02d:%02d", hours,  min, sec );
  }
  else if (solutions > 1)
  {
    sprintf(msgbuf,"NOT a real sudoku, there are at least %d solutions ;(\nTime spend solving= %d:%02d:%02d", solutions, hours, min, sec );
  }
  else //if (solutions ==0)
  {
    sprintf(msgbuf,"NO Solution Exists! ;(\nTime spend solving= %d:%02d:%02d", hours, min, sec );
  }
  fl_message(msgbuf);
}//solutions

void
Sudoku::printGame(char* filename)
{
	//char hintPuzzle[9] ={'-','-','-','-','v','-','-','-','-'};
	//char hintValue[9]  ={'/','-','\\','|','v','|','\\','_','/'};

	//char hintPuzzle[9] ={'=','=','=','=','v','=','=','=','='};
  //char hintPuzzle[9] ={'X','=','X','=','v','=','X','=','X'};
  //char hintValue[9]  ={'/','-','\\','|','v','|','\\','_','/'};
  char hintPuzzle[9] ={'/','-','\\','|','v','|','\\','_','/'};
	char hintValue[9]  ={'-','-','-','|','v','|','-','-','-'};

  ofstream myfile;
  myfile.open (filename);
 
  if (myfile.is_open())
  {
  	/*asci grid to start with for checking  * /
    for (int i=0; i<9;i++) 
    { 
      for (int j=0;j<9;j++) 
      {
        SudokuCell *tempcell;
        tempcell = grid_cells_[i][j];
        if (tempcell->readonlyGet())
        {
          myfile << tempcell->valueGet();
        }
        else
        {
          myfile << ".";
        }
      }// j
      myfile << "\n";
    }// i
    myfile << "\n";
    myfile << "\n";
    / * -------------------------------------- */
    /*asci BIG grid */
    //init bigGrid
    char bigGrid[37][37];
    memset(bigGrid, '.', sizeof(bigGrid));
    
    for (int ii=0; ii<37;ii++)
    {
      for (int jj=0;jj<37;jj++)
    	{
    		if (ii % 4 == 0) //row border
    		{
          if (ii % 12 == 0) 
            bigGrid[ii][jj] = '='; //subgrid
          else
            bigGrid[ii][jj] = '-'; //row

    			if (jj % 4 == 0) // block border
    			{
            if (jj % 12 == 0) 
              bigGrid[ii][jj] = '#'; //subgrid
            else
              bigGrid[ii][jj] = '+'; //row
    			}
    		}//ii%4
    		else
    		{
    			if (jj % 4 == 0) // block border
    			{
            if (jj % 12 == 0) 
              bigGrid[ii][jj] = '"'; //subgrid
            else
              bigGrid[ii][jj] = '|'; //row
    			}
    		}
    	}//jj
    }//ii
  
    //fill the biggrid based on the cells
    for (int i=0; i<9;i++)
    {
      for (int j=0;j<9;j++)
      {
        SudokuCell *tempcell;
        tempcell = grid_cells_[i][j];
        //print this cell
        
        if (tempcell->readonlyGet()) //puzzle value
        {
        	for (int it = 0; it < 9; it++)
 	        	bigGrid[i*4 + (it/3) +1][j*4 + (it%3) + 1]= hintPuzzle[it];
          bigGrid[i*4+2][j*4+2]= '0' + tempcell->valueGet();
        }
        else if (tempcell->valueGet())  //user value
        {
        	for (int it = 0; it < 9; it++)
 	        	bigGrid[i*4 + (it/3) +1][j*4 + (it%3) + 1]= hintValue[it];
          bigGrid[i*4+2][j*4+2]= '0' + tempcell->valueGet();
        }
        else //no value print the hints
        {
        	for (int it = 0; it < 9; it++)
 	        	bigGrid[i*4 + (it/3) +1][j*4 + (it%3) + 1]=  tempcell->hintsGetChar(it + 1);
        }
      }
    }
    
    //print bigGrid
    for (int ii=0; ii<37;ii++)
    {
      for (int jj=0;jj<37;jj++)
      {
        myfile << bigGrid[ii][jj];
        myfile << " ";
      }
      myfile << "\n";
    }
    myfile << "\n";
    myfile << "\n";
    
    /* -------------------------------------- */
    /*asci BIGER grid */
    //init bigerGrid
    char bigerGrid[73][73];
    memset(bigerGrid, '.', sizeof(bigerGrid));
    
    char hintBigPuzzle[49] ={ '+','-','-','-','-','-','+',
												    	'|','+',' ',' ',' ','+','|',
												    	'|',' ','+',' ','+',' ','|',
												    	'|',' ',' ','V',' ',' ','|',
												    	'|',' ','+',' ','+',' ','|',
												    	'|','+',' ',' ',' ','+','|',
												    	'+','-','-','-','-','-','+'};
												    	
	  char hintBigValue[49]  ={ '/','-','-','-','-','-','\\',
												    	'|',' ',' ',' ',' ',' ','|',
												    	'|',' ',' ',' ',' ',' ','|',
												    	'|',' ',' ','V',' ',' ','|',
												    	'|',' ',' ',' ',' ',' ','|',
												    	'|',' ',' ',' ',' ',' ','|',
												    	'\\','-','-','-','-','-','/'};
	
    for (int ii=0; ii<73;ii++)
    {
      for (int jj=0;jj<73;jj++)
    	{
    		if ((ii % 2 == 0) && (jj % 2)) bigerGrid[ii][jj] = ' ';  //clean up
        if (ii % 2 == 1) bigerGrid[ii][jj] = ' ';  //clean up
    		if (ii % 8 == 0) //row border
    		{
          if (ii % 24 == 0) 
            bigerGrid[ii][jj] = '='; //subgrid
          else
            bigerGrid[ii][jj] = '-'; //row

    			if (jj % 8 == 0) // block border
    			{
            if (jj % 24 == 0) 
              bigerGrid[ii][jj] = '#'; //subgrid
            else
              bigerGrid[ii][jj] = '+'; //row
    			}
    		}//ii%4
    		else
    		{
    			if (jj % 8 == 0) // block border
    			{
            if (jj % 24 == 0) 
              bigerGrid[ii][jj] = '"'; //subgrid
            else
              bigerGrid[ii][jj] = '|'; //row
    			}
    		}
    	}//jj
    }//ii
    
    //fill the biggergrid based on the cells
    for (int i=0; i<9;i++)
    {
      for (int j=0;j<9;j++)
      {
        SudokuCell *tempcell;
        tempcell = grid_cells_[i][j];
        
//if (tempcell->readonlyGet()){bigerGrid[i*8+4][j*8+4]= '0' + tempcell->valueGet();}
        
        if (tempcell->readonlyGet()) //puzzle value
        {
          for (int it = 0; it < 49; it++)
   	       	bigerGrid[i*8 + (it/7) +1][j*8 + (it%7) + 1]= hintBigPuzzle[it];
          bigerGrid[i*8+4][j*8+4]= '0' + tempcell->valueGet();
        }
        else if (tempcell->valueGet())  //user value
        {
        	for (int it = 0; it < 49; it++)
 	        	bigerGrid[i*8 + (it/7) +1][j*8 + (it%7) + 1]= hintBigValue[it];
          bigerGrid[i*8+4][j*8+4]= '0' + tempcell->valueGet();
        }
        else //no value print the hints
        {
        	for (int it = 0; it < 9; it++)
 	        	bigerGrid[i*8 + (it/3)*2 + 2][j*8 + (it%3)*2 + 2]=  tempcell->hintsGetChar(it + 1);
        }//puzlle/value/hints

      }// j
    }// i
 
    //print bigerRGrid
    for (int ii=0; ii<73;ii++)
    {
      for (int jj=0;jj<73;jj++)
      {
        myfile << bigerGrid[ii][jj];
      }
      myfile << "\n";
    }
    myfile << "\n";
    myfile << "\n";

    myfile << "The End, That's all Folks!\n";
    myfile.close();
  }// if open
}// printGame

// ---solver solutions helpers------------------------------------------------
// Recursive backtracking routine
// performance improvement by checking from time to time the grid so fare,
//   and quiting wrong paths immediately
void
Sudoku::solveGrid(int startValue, int* solutionsPtr)
{
  if (startValue == 81) //solved (0-80) dus alles is ingevuld neem ik aan
  {
    //check the filled in version with the check routine.
    if (checkGrid())
    {
      //found a solution!
      *solutionsPtr = *solutionsPtr + 1;

      printf("Solution [%d] checks:%d => ", *solutionsPtr, solvechecks);
      printGrid();
      fflush(stdout);
      if (*solutionsPtr == 2)  //or 5 or ...
      {
        solutionsStopped = true;
      }

      //no solution and first found capture this one
      if ((!hassolution_) && (*solutionsPtr == 1))
      {
        //save the solution to the the puzzle
        //printf("solveGrid SAVE solution to the puzzle!\n");
        for (int i=0; i<9 ; i++)
          for (int j=0; j<9 ; j++)
          {
            int value = testGrid[i][j][0];
            grid_values_[i][j] = value;
          }
        hassolution_ = 1;
        prefs_.set("hassolution", hassolution_);
      }//save first solution
    }//found a solution
  }//startValue 81 => end of grid
  else
  {
    //still parts of grid to go
    int i = startValue / 9;
    int j = startValue % 9;

    //Performance improvement, quiting all possible following solutions,
    //  because the existing part is already wrong
    //TODO improve by performing this check when smart
    //     (based on number of empty cells and hints or other smart ideas)
    if (j==0) // or   startValue % ANYVALUE
    {
      if (!checkGrid())
      {
        //printf("Illegal partial sudoku at (%d,%d)\n", i,j);
        return; //wrong path
      }
    }

    if (!testGrid[i][j][10])
    {
      //foreach hint in this cell do as value
      for (int h=1; h<10 && !solutionsStopped; h++)
      {
        if (testGrid[i][j][h])
        {
          testGrid[i][j][0] = h;
          //printf("choose (%d,%d) => %d\n", i,j ,h);
          //printf("HINTS => value  ");
          //printGridcell(i,j);
          solveGrid (startValue+1, solutionsPtr);
        }
      }
      testGrid[i][j][0] = 0;
      //restore the value for next trial,
      // needed due to the intermediate checks which need a 0 for empty cells
    }
    else
    {
      // has no hints
      solveGrid(startValue+1, solutionsPtr);
    }
  }//check or continue
}// solveGrid

//Check the grid for doubles in row/col/box
//  based on the check_game routine but ignores empty cells
int Sudoku::checkGrid()
{
  solvechecks++;
  //if( solvechecks%10000==0 ) printf("%ld checkGrid\n", solvechecks ); fflush ( stdout );
  int result = 0; // 1: OK 0: wrong

  bool correct = true;
  int i, j, k;

  // Check the game for right/wrong answers...
  for (i = 0; i < 9; i ++)
  {
    for (j = 0; j < 9; j ++)
    {
      int val = testGrid[i][j][0];

      if (val)
      {
        for (k = 0; k < 9; k ++)
        {
          int boxi = (i/3)*3 + (k%3);
          int boxj = (j/3)*3 + k/3;

          if ((i != k && testGrid[k][j][0] == val) ||
              (j != k && testGrid[i][k][0] == val) ||
              (boxi != i && boxj != j && testGrid[boxi][boxj][0] == val) )
          {
            break;
          }
        }
        if (k < 9)
        {
          correct = false;
        }
      }//val
    }//j
  }//i
  if (correct)// Success!
  {
    result = 1; // 1: OK 0: wrong
  }
  return result;
}//checkGrid

// Print the grid routine
void
Sudoku::printGrid()
{
  for (int i=0; i<9;i++)
    for (int j=0;j<9;j++)
    {
      printf("%d", testGrid[i][j][0] );
    }
  printf("\n");
}//printGrid

//print the gridcell
//CELL (i,j): value, hint1, 2, 3, 4, 5, 6, 7, 8, 9, readonly,;
void
Sudoku::printGridcell(int printi, int printj)
{
  printf("CELL (%d,%d):",printi,printj);
  for ( int i=0; i<=10; i++)
  {
    printf("%d, ",testGrid[printi][printj][i] );
  }
  printf(";\n");
}//printGridcell
// ---helper------------------------------------------------------------------
// Load the game from saved preferences...
void
Sudoku::load_pref_game()
{
  // Load the current values and state of each grid...
  memset(grid_values_, 0, sizeof(grid_values_));

  bool solved = true;

  bool hasHints = false; //only autohints if no hints stored, don't spoil the users bookkeeping
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      char name[255];
      int val;

      SudokuCell *cell = grid_cells_[i][j];

      sprintf(name, "value%d.%d", i, j);
      if (!prefs_.get(name, val, 0))
      {
        i = 9;
        grid_values_[0][0] = 0;
        break;
      }

      grid_values_[i][j] = val;

      sprintf(name, "state%d.%d", i, j);
      prefs_.get(name, val, 0);
      cell->valueSet(val);

      sprintf(name, "readonly%d.%d", i, j);
      prefs_.get(name, val, 0);
      cell->readonlySet(val);

      if (val)
        cell->color(FL_GRAY);
      else
      {
        cell->color(FL_LIGHT3);
        solved = false;
      }

      for (int k = 0; k < 10; k ++)
      {
        sprintf(name, "test%d%d.%d", k, i, j);
        prefs_.get(name, val, 0);
        cell->hintsSet(val);
        if ((val == k) && (k!=0))
        {
          hasHints = true;
        }
      }
    }//j
  }//i
  // If we didn't load any values or the last game was solved, then
  // create a new game automatically...
  //TEMP remove to allow continueing imported games which are not complete

  prefs_.get("difficulty", difficulty_, 0);

  prefs_.get("hassound", hassound_, 1);  //1 is default???
  sound_->soundon(hassound_);

  prefs_.get("autohints", autohints_, 0);  //0 is default???

  prefs_.get("autounique", autounique_, 0);  //0 is default???

  prefs_.get("promoteunique", promoteunique_, 0);  //0 is default???

  prefs_.get("hassolution", hassolution_, 1);  //1 is default??? (diff used)

  prefs_.get("imported", imported_, 0);  //0 is default??? (diff used)

  //TODO do we want this??
  if (solved)
  {
    new_game(time(NULL));
  }
  else
  {
    check_game(false);
  }

  if (!hasHints && autohints_)
  {
    //only autohints if no hints stored, don't spoil the users bookkeeping
    hints();
  }

  if (autounique_) showuniquehints();
}//load_game

// Save the current game state the the  preferences...
void
Sudoku::save_pref_game()
{
  for (int i = 0; i < 9; i ++)
  {
    for (int j = 0; j < 9; j ++)
    {
      char name[255];
      SudokuCell *cell = grid_cells_[i][j];

      sprintf(name, "value%d.%d", i, j);
      prefs_.set(name, grid_values_[i][j]);

      sprintf(name, "state%d.%d", i, j);
      prefs_.set(name, cell->valueGet());

      sprintf(name, "readonly%d.%d", i, j);
      prefs_.set(name, cell->readonlyGet());

      for (int k = 0; k < 10; k ++)
      {
        sprintf(name, "test%d%d.%d", k, i, j);
        prefs_.set(name, cell->hintsGet(k));
      }
    }//j
  }//i
}//save_game

// Return the next available value for a cell...
//int
//Sudoku::next_value(SudokuCell *c)
int
Sudoku::next_value(int row, int col)
{	
  int	i, j, k, m;
  SudokuCell *c;

  //set index to top of box??
  i = row - row % 3;
  j = col - col % 3;

  //the availabble numbers inited at 0
  int numbers[9];
  memset(numbers, 0, sizeof(numbers));

  //loop box
  for (k = 0; k < 3; k ++)
  {
    for (m = 0; m < 3; m ++)
    {
      c = grid_cells_[i + k][j + m];
      if (c->valueGet()) numbers[c->valueGet() - 1] = 1;
    }
  }
  //loop row & columns
  for (k = 0; k < 9; k ++)
  {
    //loop col   
  	c = grid_cells_[k][col];
    if (c->valueGet()) 
    { 
    	numbers[c->valueGet() - 1] = 1;
    }     
    //row
  	c = grid_cells_[row][k];
    if (c->valueGet())
    { 
    	numbers[c->valueGet() - 1] = 1; 	
    }
  }
 
  //loop results till first empty one
  for (i = 0; i < 9; i ++)
    if (!numbers[i]) return i + 1;

  //nothing found return
  return 1;
}// next_value

//Helper function for import_game
// returns the value of a position in a character string
// default is 0, even if the string is to short
//TODO improve this with stringlen / etc.
int
Sudoku::getNextValue(char* importString, int teller )
{
  int result = 0;
  char tempchar = importString[teller];
  if (tempchar =='\0')
  {
    import_stopped = true;
  }
  if (!import_stopped )
  {
    if (tempchar <= '0')
    {
      result = 0; //printf("  te klein\n");
    }
    else if (tempchar > '9')
    {
      result = 0; //printf("  te groot\n");
    }
    else
    {
      result = tempchar - '0'; //printf("  correct [%d]\n", result);
    }
  }
  return result;
}//getNextValue

//helper function for show unique hints
void Sudoku::testcell(SudokuCell * cell, int* hints, int index)
{
  char hintvalue;
  if (!cell->valueGet() ) //look at the hints
  {
    int singleHint	= 0; //0 not one yet, 1 - 9 the hint found, 10 more than one hint

    //loop all 9 hints
    for (int h=1;h<=9;h++)
    {
      hintvalue = cell->hintsGet(h);  //[0,9] filled 1-9

      if (hintvalue) //found hint h in this cell : bookkeep
      {
        if (!singleHint)
          singleHint = hintvalue;
        else
          if (singleHint < 10) singleHint = 10;

        int currentKept = hints[h];
        if (currentKept == -1)//not yet registered
        {
          //none yet
          hints[h] = index; //[1,9]
        }
        else if (currentKept >= 0)
        {
          //already one exists
          hints[h] = -2;
        }
        //else { more than one exists curentKept should be -2	}
      }// there is a hint in this cell
      // else { no hintvalue }
    }//h loop hints
    if (singleHint > 0 && singleHint < 10)
    {
      //a single hint is found and thus unique for this cell instead for a row or column :)
      cellHint[cell->myI][cell->myJ] = singleHint;
    }
  }//only look at empty cells
  //else { values exist printf("cell (%d,%d) has value %d\n", cell->myI, cell->myJ, cell->value());	}
}//testcell

//helper function for testcell
void Sudoku::printHints(int *hints)
{
  printf("hints =[");
  for (int i = 0; i<9 ; i++)
  {
    printf("%d, ", hints[i]);
  }
  printf("]\n");
} // printHints array

//lets the cell update the last focused cell info of the game
void
Sudoku::setFocusCell(int I,int J)
{
  lastFocusI = I;
  lastFocusJ = J;
}//setFocusCell

int
Sudoku::getautohints()
{
  return autohints_;
}

int
Sudoku::getautounique()
{
  return autounique_;
}

int
Sudoku::getpromoteunique()
{
  return promoteunique_;
}

// ---FLTK UI related---------------------------------------------------------
// Resize the window...
void
Sudoku::resize(int X, int Y, int W, int H)
{
  // Resize the window...
  Fl_Window::resize(X, Y, W, H);

  // Save the new window geometry...
  prefs_.set("x", X);
  prefs_.set("y", Y);
  prefs_.set("width", W);
  prefs_.set("height", H);
}// resize

// Set title of window...
void
Sudoku::set_title()
{
  static const char * const titles[] =
    {
      "Sudoku - Easy",
      "Sudoku - Medium",
      "Sudoku - Hard",
      "Sudoku - Impossible",
      "Sudoku - !!Imported!!"
    };
  int labelitem = difficulty_;
  if (imported_) labelitem = 4;
  label(titles[labelitem]);
}//set_title
// ---------------------------------------------------------------------------
/*
 * print game to file for paper use:
 * #===#===#===#===#
 * +123|/ \|///+
 * +456| 7 |/3/+
 * +789|\ /|///+
 * #---+---+---+---#
 * +123|/ \|///+
 * +456|
 * +789|

 * #===
 *
 */
